from abc import ABC, abstractmethod
from attr import dataclass
import requests

from constants import REPOSITORY_URL
from instance import Instance, InstanceData


@dataclass
class ProjectMetadata:
    name: str
    path: str


class Project(ABC):
    def __init__(self, network, metadata: ProjectMetadata) -> None:
        super().__init__()
        self.network = network
        self._metadata = metadata

    def get_instance(self, aliasOrAddress: str):
        if self.network.w3.is_address(aliasOrAddress):
            address = aliasOrAddress
            return self.createInstance(address)
        else:
            alias = aliasOrAddress
            url = self.get_instance_data_url(alias)
            response = requests.get(url)
            if (response.status_code != 200):
                raise ValueError(f"Can't download data for instance {alias}")
            else:
                data = InstanceData.from_json(response.text)
                if len(data.data.values()) != 1:
                    raise ValueError(f'Error during parsing data for {alias}')
                return self.create_instance(list(data.data.values())[0])

    def download_abi_file(self, version: str):
        response = requests.get(self.get_abi_url(version))
        return response.text

    def get_abi_url(self, version: str):
        return f'{self.githubRepo}releases/download/{version}/{self.get_abi_filename(version)}'

    @abstractmethod
    def get_abi_filename(self, version: str) -> str:
        pass

    def get_instance_data_url(self, alias: str):
        from network import ListedNetwork
        if isinstance(self.network, ListedNetwork):
            return f'{REPOSITORY_URL}{self.network.path}/{self._metadata.path}/{alias}.json'
        else:
            raise ValueError('Network is unknown')

    @abstractmethod
    def create_instance(self, address: str) -> Instance:
        pass



from web3 import Web3
from web3.providers import BaseProvider

from projects.project_factory import project_factory

class Network:
    def __init__(self, skale_contracts, provider: BaseProvider):
        self.w3 = Web3(provider)
        self._skaleContracts = skale_contracts

    def get_project(self, name: str):
        return project_factory.create(self, name)

class ListedNetwork(Network):
    def __init__(self, skale_contracts, provider: BaseProvider, path: str):
        super().__init__(skale_contracts, provider)
        self.path = path



from abc import ABC, abstractmethod
from attr import dataclass
import json


@dataclass
class InstanceData:
    data: dict[str, str]
    @classmethod
    def from_json(cls, data: str):
        return cls(data=json.loads(data))


class Instance(ABC):
    def __init__(self, project, address: str) -> None:
        self._project = project
        self._version = None
        self._abi = None
        self.address = address

    def get_w3(self):
        return self._project.network.w3

    def get_version(self):
        if self._version is None:
            self._version = self._get_version()
            if not '-' in self._version:
                self._version = self._version + '-stable.0'
        return self._version

    def get_abi(self):
        if self._abi is None:
            self._abi = json.loads(self._project.download_abi_file(self.version))
        return self._abi

    abi = property(get_abi, None)
    version = property(get_version, None)
    w3 = property(get_w3, None)

    @abstractmethod
    def get_contract_address(name: str) -> str:
        pass

    def get_contract(self, name: str):
        address = self.get_contract_address(name)
        return self.w3.eth.contract(address=address, abi=self.abi[name])

    # protected

    @abstractmethod
    def _get_version() -> str:
        pass



from web3 import Web3
from web3.providers import BaseProvider

from metadata import Metadata
from network import Network, ListedNetwork


class SkaleContracts:
    def __init__(self):
        self.metadata = Metadata()

    def get_network_by_provider(self, provider: BaseProvider) -> Network:
        w3 = Web3(provider)
        chain_id = w3.eth.chain_id
        networkMetadata = self.metadata.get_network_by_chain_id(chain_id)
        if networkMetadata is None:
            return Network(self, provider)
        else:
            return ListedNetwork(self, provider, networkMetadata.path)
